/****************************************************************************
 * Grafik-LCD Treiber mit VRAM.
 *
 * (C) Andreas Kaiser 2008.
 *
 * Die Software darf frei kopiert und verndert werden, solange sie
 * nicht ohne meine ausdrckliche Erlaubnis fr kommerzielle Zwecke
 * eingesetzt wird.
 ****************************************************************************/

#include "config.h"
#include "vram.h"
#include "lcd.h"
#include "gbase.h"

static uint8_t		width, height;
static g_font_t *	font;
static g_font_t		softfont = { 8, 12, NULL };

static g_y		g_curx;
static g_x		g_cury;

#define SOFT_BASE	(VRAM_COLS - G_COLS/LCD_BITS)

static void set_cmode (void);

//-----------------------------------------------------------------------------

bool
g_gpos(g_x x, g_y y)
{
    //PRINTF("glcd_gpos(%d,%d)\n", x, y);
    if (x >= G_COLS || y >= G_ROWS)
	return false;
    g_cury = y;
    g_curx = x;
    return true;
}

//-----------------------------------------------------------------------------

extern g_font_t	font_8_6;
extern g_font_t	font_10_8;
extern g_font_t	font_12_8;
extern g_font_t	font_48_32;

static g_font_t *fonts[] = {
#if GLCD_FONT_8_6
	&font_8_6,
#endif
#if GLCD_FONT_10_8
	&font_10_8,
#endif
#if GLCD_FONT_12_8
	&font_12_8,
#endif
#if GLCD_FONT_48_32
	&font_48_32,
#endif
	0
};

bool
g_hardfont(uint8_t w, uint8_t h)
{
    for (uint8_t i = 0; fonts[i]; ++i) {
	if (fonts[i]->height == h && fonts[i]->width == w) {
	    font = fonts[i];
	    height = h;
	    width  = w;
	    set_cmode();
	    return true;
	}
    }
    return false;
}

bool
g_softfont(uint8_t w, uint8_t h)
{
    uint16_t nbits = (VRAM_COLS - G_COLS/LCD_BITS) * VRAM_BITS;
    if (umul8x8(height, width) > nbits)
	return false;
    font = &softfont;
    font->height = height = h;
    font->width  = width  = h;
    set_cmode();
    return true;
}

bool
g_softchar(uint8_t ch, uint8_t off, uint8_t data)
{
    uint16_t nbits = (VRAM_COLS - G_COLS/LCD_BITS) * VRAM_BITS;
    uint16_t nbytes = nbits / 8;
    if (off >= nbytes)
	return false;

    if (VRAM_BITS == 4) {
	vram_col_t col = SOFT_BASE + 2*off;
	vram_write(ch, col+0, 0xF0, data);
	vram_write(ch, col+1, 0xF0, data << 4);
    } else {
	vram_col_t col = SOFT_BASE + off;
	vram_write(ch, col, 0xFF, data);
    }

    return true;
}

//-----------------------------------------------------------------------------

static uint8_t
softchar(uint8_t ch, uint8_t ccol)
{
    uint8_t data;
    if (VRAM_BITS == 4) {
	data  = vram_read(ch, SOFT_BASE + 2*ccol+1) & 0xF0;
	data |= vram_read(ch, SOFT_BASE + 2*ccol+0) >> 4;
    } else {
	data = vram_read(ch, SOFT_BASE + ccol);
    }
    return data;
}

static void
chardot(g_x col, g_y row, g_op op, uint8_t data)
{
    if (op == G_Replace)
	op = (data & 0x80) ? G_Set : G_Clear;
    else if (!(data & 0x80))
	return;
    g_dot(col, row, op);
}

bool
g_gchar(uint8_t ch, g_op op, int8_t angle)
{
    prog_uchar *ptr = 0;
    if (font != &softfont) {
	uint8_t nbytes = (umul8x8(width, height) + 7) / 8;
	ptr = font->data + umul8x8(ch, nbytes);
    }

    if (width == 8
     && (g_curx % 8) == 0
     && op == G_Replace
     && angle == 0) {

	vram_row_t vrow = g_cury;
	vram_col_t vcol = g_curx / LCD_BITS;

	for (uint8_t crow = 0; crow < height; ++crow) {
	    uint8_t data = ptr ? pgm_read_byte(ptr++) : softchar(ch, crow);
	    if (LCD_BITS == 4) {
		g_write(vrow+crow, vcol+0, G_Replace, 0xF0, data);
		g_write(vrow+crow, vcol+1, G_Replace, 0xF0, data << 4);
	    } else {
		g_write(vrow+crow, vcol, G_Replace, 0xFF, data);
	    }
	}

	if (g_curx + 8 < G_COLS)
	    g_curx += 8;

    } else {

	int8_t rstep, cstep;
	switch (angle) {
	case 0:
	    rstep = 0;
	    cstep = 1;
	    break;
	case 45:
	    rstep = 1;
	    cstep = 1;
	    break;
	case 90:
	    rstep = 1;
	    cstep = 0;
	    break;
	case -45:
	    rstep = -1;
	    cstep = 1;
	    break;
	case -90:
	    rstep = -1;
	    cstep = 0;
	    break;
	default:
	    return false;
	}

	uint8_t data = 0;
	uint8_t bit  = 0;
	for (uint8_t crow = 0; crow < height; ++crow) {
	    g_y drow = g_cury + smul8x8(crow, cstep);
	    g_y dcol = g_curx + smul8x8(crow, rstep);
	    for (uint8_t ccol = 0; ccol < width; ++ccol) {
		if ((bit++ % 8) == 0)
		    data = ptr ? pgm_read_byte(ptr++) : softchar(ch, ccol);
		chardot(dcol, drow, op, data);
		data <<= 1;
		drow -= rstep;
		dcol += cstep;
	    }
	}

	g_gpos(g_curx + smul8x8(cstep, width),
	       g_cury - smul8x8(rstep, width));
    }
    return true;
}

void
g_gtext(const char *s, g_op op, int8_t angle)
{
    while (*s) {
	g_gchar(*s++, op, angle);
    }
}

void
g_gtext_p(const prog_char *s, g_op op, int8_t angle)
{
    for (char ch; (ch = pgm_read_byte(s)) != 0; ++s) {
	g_gchar(ch, op, angle);
    }
}

//-----------------------------------------------------------------------------
// char mode

uint8_t		c_row, c_col;
uint8_t		c_rows, c_cols;

static void
set_cmode(void)
{
    c_rows = G_ROWS / height;
    c_cols = G_COLS / width;
}

bool
g_cpos(uint8_t ccol, uint8_t crow)
{
    if (crow >= c_rows || c_cols >= c_cols)
	return false;
    c_row = crow;
    c_col = ccol;
    return true;
}

static void
g_cfill(uint8_t crow, uint8_t ccol, uint8_t n)
{
    g_y y = umul8x8(crow, height);
    g_x x = umul8x8(ccol, width);
    g_x nbits = (ccol+n >= c_cols) ? G_COLS - ccol : umul8x8(n, width);
    g_fill(x, y, x+nbits-1, y+height-1, G_Clear);
}

static void
g_newline(void)
{
    if (c_row == c_rows-1)
	lcd_set_home(lcd_home + height);
    else
	++c_row;
    c_col = 0;
    g_cfill(c_row, 0, c_cols);
}

void
g_char(uint8_t ch)
{
    //PRINTF("glcd_char(%02X)\n", ch);
    switch (ch) {

    case '\n':
	g_newline();
	break;

    case '\r':
	c_col = 0;
	break;

    case '\t':
	c_col += 8 - (c_col % 8);
	break;

    case '\f':
	g_clear();
	c_row = 0;
	c_col = 0;
	break;

    default:
	if (c_col >= c_cols)
	    g_newline();
	g_gpos(umul8x8(c_col, width), umul8x8(c_row, height));
	g_gchar(ch, G_Replace, 0);
	++c_col;
	break;
    }
}

void
g_text(const char *s)
{
    while (*s) {
	g_char(*s++);
    }
}

void
g_text_p(const prog_char *s)
{
    for (char ch; (ch = pgm_read_byte(s)) != 0; ++s) {
	g_char(ch);
    }
}
